/*	===========================================================================
 UPC - EETAC - CSD - DEE - http://digsys.upc.edu

 * LAB6: Analogue inputs, analogue sensors
 * Documentation location: https://digsys.upc.edu/csd/DEE/lab06/lab06.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P
 * Similar circuit: 	https://digsys.upc.edu/csd/P12/Temp/Temp.html

Planning: 

- Copy and adapt the previous project LAB5: FSM with interrupts
				
- Prepare the hardware, so that adapt the Arduino board to this 
particular application. For instance:

    - A/D function. Measure temperature 

- Prepare the software. Hardware/software diagram. State diagram

    - Real-time measurement using millis() 
	- Sampling period flag. 
	- Process and transmit temperatures
=============================================================================*/


/*	===========================================================================
	Definitions
	======================================================================== */

const char pin_AD0	= A0;  // Analogue inputs
const char pin_AD1  = A1;
const char pin_AD2  = A2;

// Start-stop push-button. Pin PD3 is also the microcontroller's external INT1
const char INT1_pin_ST_BUTTON = 3;  // 

// LEDs for indicating the current state 
const char pin_LED2 = 9; 
const char pin_LED1 = 6; 
const char pin_LED0 = 5; 

// LED for generating the 1 Hz squared waveform when acquiring
const char pin_SLED = 17; 

 
// FSM list of states
// FSM list of states 
const char IDLE           = 'A'; 
const char SET_ACQ_TIMER  = 'B';
const char Do_NOTHING     = 'C';
const char ACQUIRE        = 'D';
const char TEMP_MEAS      = 'E';
const char TRANSMIT       = 'F';
const char STOP_ACQ_TIMER = 'G';


// Default timer period to sample analogue inputs 
const unsigned long SAMPLING_PERIOD = 500; // 500 ms   (2 Hz sampling frequency)

/*	===========================================================================
	Global RAM variables
	======================================================================== */


char   	var_S; 			// For representing the state in binary
char   	var_SLED; 		// Squared waveform when running

char 	var_current_state ;    // This is the key variable: FSM state memory

char 	var_SP_flag; 	// Sample period flag 
char 	var_ST_flag;    // It will be set when one click detected on interrupt

char 	var_USART_flag; // For writing or sending ASCII messages only when 
						// new information

// Voltage variables
static unsigned int var_Vs0;   
static unsigned int var_Vs1; 
static unsigned int var_Vs2;   

// Temperature variables
static float var_Ts0; 
static float var_Ts1;   
static float var_Ts2;  



// For timing the square waveform: 
static unsigned long var_previous_time;   
static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed;   //  time elapsed since last flag

static unsigned int var_sampling_period; 


/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
// Configuring digital output pins for indicating the current state in binary
// for debugging purposes
	pinMode(pin_LED2, OUTPUT);  
	pinMode(pin_LED1, OUTPUT);	
	pinMode(pin_LED0, OUTPUT);	

// Output where to connect the sampling LED that will toggle every Ts	
	pinMode(pin_SLED, OUTPUT);	

// Configuring digital input pin for reading VM (only in design step #2)
	// pinMode(pin_VM, INPUT);  

// This is the special pin to be an external interrupt source INT1:
	pinMode(INT1_pin_ST_BUTTON, INPUT_PULLUP); 
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
 //Initialise the FSM and the timer 
	var_current_state = IDLE;	
	var_previous_time = 0; 	 

		// Serial monitor ASCII message:
	Serial.println("Project LAB5 - DEE");
	Serial.println("Temperature acquisition, sampling three analogue inputs"); 
	Serial.println("System IDLE: waiting for ST click to start"); 

//	---------------------------------------------------------------------------
// This is the interrupt handler. Read the Arduino documentation at: 
// https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/
// attachInterrupt(digitalPinToInterrupt(pin), ISR, mode) (recommended)
	
	attachInterrupt(digitalPinToInterrupt(INT1_pin_ST_BUTTON), interrupt_service_routine_INT1, FALLING);
	// Interrupt when detecting a falling edge to start and stop operations
}

/*	===========================================================================
	Infinite loop to run the FSM
	======================================================================== */
void loop() {       
	state_logic();    // calculating the next state to go
	output_logic();		// calculating outputs 
}

/*	===========================================================================
	State logic to determine where to go next
	======================================================================== */
void state_logic(void){
	read_inputs(); 		// capture the switch voltage values into RAM variables
	update_time_flag();    //  to keep track of real time and be able to 
							// acquire signals at sampling frequency

	switch (var_current_state) {
	case IDLE:	// Do nothing, wait for start signal
		if (var_ST_flag == 1) {	  // start the process
			  var_current_state = SET_ACQ_TIMER;
			  var_ST_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = IDLE;     
		break;

	case SET_ACQ_TIMER:	// Do nothing, wait for start signal
		var_current_state = Do_NOTHING;       
		break;

	case Do_NOTHING:	// Wait for sampling period
		if (var_SP_flag == 1) {	  // start acquiring 
			var_current_state = ACQUIRE;
			var_SP_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;

	case ACQUIRE:
		var_current_state = TEMP_MEAS;     
		break;

	case TEMP_MEAS:	
		var_current_state = TRANSMIT;  
		break;

	case TRANSMIT:
		if (var_ST_flag == 1) {	  // stop acquiring 
			var_current_state = STOP_ACQ_TIMER;
			var_ST_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;
		
	case STOP_ACQ_TIMER:	// go to Idle to wait for another Start pulse
		var_current_state = IDLE;     
		break;

	}	
}

/*	===========================================================================
	Output logic 
	======================================================================== */
void output_logic(void){
int var_buff;
	
	switch (var_current_state) {
	case IDLE:
		var_SLED = 0;
		var_S = 0;		
	
		break;

	case SET_ACQ_TIMER:	
		var_SLED = 0;
		var_S = 1;	
		var_sampling_period = SAMPLING_PERIOD; 
		break;

	case Do_NOTHING: 
		var_S = 2;	
		break;	


	case ACQUIRE: 
		// Toggle operation with a XOR simply to indicate running acquisition
		var_SLED = var_SLED^0b00000001; 
		var_S = 3;	
		
			// Three consecutive analogue to digital conversions
		var_Vs0 = analogRead(pin_AD0); // read V_AD0 voltage
		var_Vs1 = analogRead(pin_AD1); // read V_AD1 voltage
		var_Vs2 = analogRead(pin_AD2); // read V_AD2 voltage

		break;	


	case TEMP_MEAS:	// measurement equation
		var_S = 4;	
		var_Ts0 = (var_Vs0 - 307.2)/10.24; 
		var_Ts1 = (var_Vs1 - 307.2)/10.24; 	
		var_Ts2 = (var_Vs2 - 307.2)/10.24; 

	// Rounding a float with only one decimal digit
		var_Ts0 = var_Ts0*10 + 0.5;
		var_buff = (int)var_Ts0;
		var_Ts0 = (float)var_buff/10;

		var_Ts1 = var_Ts1*10 + 0.5;
		var_buff = (int)var_Ts1;
		var_Ts1 = (float)var_buff/10;

		var_Ts2 = var_Ts2*10 + 0.5;
		var_buff = (int)var_Ts2;
		var_Ts2 = (float)var_buff/10;

		break;
		
		
	case TRANSMIT:	// measurement equation	
		var_S = 5;	
	
		// Serial monitor:
		Serial.print("Ts0 = "); 	
		Serial.print((float)var_Ts0, 1); 		
		Serial.print(" degC     "); 

		Serial.print("Ts1 = "); 	
		Serial.print((float)var_Ts1, 1); 		
		Serial.print(" degC     "); 

		Serial.print("Ts2 = "); 	
		Serial.print((float)var_Ts2, 1); 		
		Serial.println(" degC"); 

		break;

	case STOP_ACQ_TIMER:
		var_S = 6;		
		break;
	
	}
	write_outputs(); // Convert RAM variable values into pin output voltage	
}	


/*	===========================================================================
	ISR external interrupt INT0 pin 
This function is executed when an active edge is detected at pin PD2 (INT0)
The idea is simply set the global variable
	======================================================================== */
void interrupt_service_routine_INT1 (void){

	var_ST_flag = 1;  

	// LAB experimentation on debouncing techniques: 
	//delay(5); // wait some time 

}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){


  // Do nothing now in design step #1 

 
}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
char var_buf;  // local variable for saving partial results

	var_buf = var_S & 0b00000001; //Masking the bit of interest
   if (var_buf == 1)
       digitalWrite(pin_LED0, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED0, HIGH);  
	
	var_buf = (var_S & 0b00000010)>>1;
   if (var_buf == 1)
       digitalWrite(pin_LED1, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED1, HIGH);  

	var_buf = (var_S & 0b00000100)>>2;
   if (var_buf == 1)
       digitalWrite(pin_LED2, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED2, HIGH);  
	

   if (var_SLED == 1)
       digitalWrite(pin_SLED, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_SLED, HIGH);  

}



/*	===========================================================================
	Establishing the sampling frequency using internal timer
	======================================================================== */
void update_time_flag(void){

  // From internal CLK: 
    var_current_time = millis();
    var_time_elapsed =  var_current_time - var_previous_time;  

	if(var_time_elapsed  > var_sampling_period) {
			// is the flag is set is because time elapsed is sampling period
		var_SP_flag = 1;   
			// to start another sampling period cycle
		var_previous_time = var_current_time; 
	}
}
